#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#


class ExpectedContentException(RuntimeError):
    def __init__(self, index, size):
        self.index = index
        self.size = size


class MissingDomainException(RuntimeError):
    def __init__(self, name):
        self.name = name


class MissingEnrichmentException(RuntimeError):
    def __init__(self, name):
        self.name = name


class MissingMetadataException(RuntimeError):
    def __init__(self, key):
        self.key = key


class MissingSourceMetadataException(MissingMetadataException):
    def __init__(self, key):
        super().__init__(key)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import json
import os
import sys
import threading
import time
import traceback
from os.path import isdir, isfile, join
from pathlib import Path
from typing import List

import pkg_resources
import requests
from deltafi.actioneventqueue import ActionEventQueue
from deltafi.domain import Event
from deltafi.exception import ExpectedContentException, MissingDomainException, MissingEnrichmentException, \
    MissingSourceMetadataException, MissingMetadataException
from deltafi.logger import get_logger
from deltafi.result import ErrorResult
from deltafi.storage import ContentService


def _coordinates():
    return {
        'groupId': os.getenv('PROJECT_GROUP'),
        'artifactId': os.getenv('PROJECT_NAME'),
        'version': os.getenv('PROJECT_VERSION')
    }


def _setup_queue(max_connections):
    redis_url = os.getenv('REDIS_URL', 'http://deltafi-redis-master:6379')
    password = os.getenv('REDIS_PASSWORD')
    return ActionEventQueue(redis_url, max_connections, password)


def _setup_content_service():
    minio_url = os.getenv('MINIO_URL', 'http://deltafi-minio:9000')
    return ContentService(minio_url,
                          os.getenv('MINIO_ACCESSKEY'),
                          os.getenv('MINIO_SECRETKEY'))


class Plugin(object):
    def __init__(self, actions: List, description: str):
        self.actions = [action() for action in actions]
        self.description = description
        self.coordinates = _coordinates()

        self.core_url = os.getenv('CORE_URL')
        self.queue = _setup_queue(len(self.actions) + 1)
        self.content_service = _setup_content_service()

        if os.getenv('ACTIONS_HOSTNAME'):
            self.hostname = os.getenv('ACTIONS_HOSTNAME')
        elif os.getenv('HOSTNAME'):
            self.hostname = os.getenv('HOSTNAME')
        elif os.getenv('COMPUTERNAME'):
            self.hostname = os.getenv('COMPUTERNAME')
        else:
            self.hostname = 'UNKNOWN'

        self.logger = get_logger()

        self.logger.info(f"Initialized ActionRunner with actions {actions}")

    def action_name(self, action):
        return f"{self.coordinates['groupId']}.{action.__class__.__name__}"

    def _action_json(self, action):
        return {
            'name': self.action_name(action),
            'description': action.description,
            'type': action.action_type.name,
            'requiresDomains': action.requires_domains,
            'requiresEnrichments': action.requires_enrichments,
            'schema': action.param_class().schema()
        }

    def _register(self):
        flows_path = str(Path(os.path.dirname(os.path.abspath(sys.argv[0]))) / 'flows')

        flow_files = []
        variables = []
        if isdir(flows_path):
            flow_files = [f for f in os.listdir(flows_path) if isfile(join(flows_path, f))]
            if 'variables.json' in flow_files:
                flow_files.remove('variables.json')
                variables = json.load(open(join(flows_path, 'variables.json')))
        else:
            self.logger.warning(f"Flows directory ({flows_path}) does not exist. No flows will be installed.")

        flows = [json.load(open(join(flows_path, f))) for f in flow_files]
        actions = [self._action_json(action) for action in self.actions]

        url = f"{self.core_url}/plugins"
        headers = {'Content-type': 'application/json'}
        registration_json = \
            {
                'pluginCoordinates': self.coordinates,
                'displayName': os.getenv('PROJECT_NAME'),
                'description': self.description,
                'actionKitVersion': pkg_resources.get_distribution('deltafi').version,
                'dependencies': [],
                'actions': actions,
                'variables': variables,
                'flowPlans': flows
            }

        self.logger.info(f"Registering plugin:\n{registration_json}")

        response = requests.post(url, headers=headers, json=registration_json)
        if not response.ok:
            self.logger.error(f"Failed to register plugin ({response.status_code}):\n{response.content}")
            exit(1)

        self.logger.info(f"Plugin registered")

    def run(self):
        self.logger.info(f"Plugin starting")
        self._register()
        for action in self.actions:
            threading.Thread(target=self._do_action, args=(action,)).start()

        threading.Thread(target=self._heartbeat).start()

        self.logger.info(f"All threads running")

        f = open("/tmp/running", "w")
        f.close()

        self.logger.info(f"Application initialization complete")

    def _heartbeat(self):
        while True:
            try:
                for action in self.actions:
                    self.queue.heartbeat(self.action_name(action))
            except Exception as e:
                self.logger.error(f"Failed to register action queue heartbeat: {e}", e)
            finally:
                time.sleep(10)

    def _do_action(self, action):
        action_logger = get_logger(self.action_name(action))

        action_logger.info(f"Listening on {self.action_name(action)}")
        while True:
            try:
                event_string = self.queue.take(self.action_name(action))
                event = Event.create(json.loads(event_string), self.hostname, self.content_service, action_logger)
                start_time = time.time()
                action_logger.debug(f"Processing event for did {event.context.did}")

                try:
                    result = action.execute(event)
                except ExpectedContentException as e:
                    result = ErrorResult(f"Action attempted to look up element {e.index + 1} (index {e.index}) from "
                                         f"content list of size {e.size}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingDomainException as e:
                    result = ErrorResult(f"Action attempted to access domain {e.name}, which does not exist",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingEnrichmentException as e:
                    result = ErrorResult(f"Action attempted to access enrichment {e.name}, which does not exist",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingSourceMetadataException as e:
                    result = ErrorResult(f"Missing ingress metadata with key {e.key}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except MissingMetadataException as e:
                    result = ErrorResult(f"Missing metadata with key {e.key}",
                                         f"{str(e)}\n{traceback.format_exc()}")
                except BaseException as e:
                    result = ErrorResult(f"Action execution {type(e)} exception", f"{str(e)}\n{traceback.format_exc()}")

                response = {
                    'did': event.context.did,
                    'action': event.context.action_name,
                    'start': start_time,
                    'stop': time.time(),
                    'type': result.result_type,
                    'metrics': [metric.json() for metric in result.metrics]
                }
                if result.result_key is not None:
                    response[result.result_key] = result.response()

                topic = 'dgs'
                if event.return_address:
                    topic += f"-{event.return_address}"
                self.queue.put(topic, json.dumps(response))
            except BaseException as e:
                action_logger.error(f"Unexpected {type(e)} error: {str(e)}\n{traceback.format_exc()}")
                time.sleep(1)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from logging import Logger
from typing import Dict, List, NamedTuple

from deltafi.storage import ContentService, ContentReference


class Content(NamedTuple):
    name: str
    content_reference: ContentReference

    def json(self):
        return {
            'name': self.name,
            'contentReference': self.content_reference.json(),
        }

    @classmethod
    def from_dict(cls, content: dict):
        if 'name' in content:
            name = content['name']
        else:
            name = None
        content_reference = ContentReference.from_dict(content['contentReference'])
        return Content(name=name,
                       content_reference=content_reference)


class Context(NamedTuple):
    did: str
    action_name: str
    ingress_flow: str
    egress_flow: str
    system: str
    hostname: str
    content_service: ContentService
    logger: Logger

    @classmethod
    def create(cls, context: dict, hostname: str, content_service: ContentService, logger: Logger):
        did = context['did']
        action_name = context['name']
        ingress_flow = context['ingressFlow']
        if 'egressFlow' in context:
            egress_flow = context['egressFlow']
        else:
            egress_flow = None
        system = context['systemName']
        return Context(did=did,
                       action_name=action_name,
                       ingress_flow=ingress_flow,
                       egress_flow=egress_flow,
                       system=system,
                       hostname=hostname,
                       content_service=content_service,
                       logger=logger)


class Domain(NamedTuple):
    name: str
    value: str
    media_type: str

    @classmethod
    def from_dict(cls, domain: dict):
        name = domain['name']
        if 'value' in domain:
            value = domain['value']
        else:
            value = None
        media_type = domain['mediaType']
        return Domain(name=name,
                      value=value,
                      media_type=media_type)


class FormattedData(NamedTuple):
    filename: str
    format_action: str
    content_reference: ContentReference
    metadata: Dict[str, str]

    @classmethod
    def from_dict(cls, formatted_data: dict):
        filename = formatted_data['filename']
        format_action = formatted_data['formatAction']
        content_reference = ContentReference.from_dict(formatted_data['contentReference'])
        metadata = formatted_data.get('metadata', {})
        return FormattedData(filename=filename,
                             format_action=format_action,
                             content_reference=content_reference,
                             metadata=metadata)


class ProtocolLayer(NamedTuple):
    action: str
    content: List[Content]
    metadata: Dict[str, str]

    @classmethod
    def from_dict(cls, layer: dict):
        action = layer['action']
        content = [Content.from_dict(item) for item in layer['content']]
        metadata = layer.get('metadata', {})
        return ProtocolLayer(action=action,
                             content=content,
                             metadata=metadata)


class SourceInfo(NamedTuple):
    filename: str
    flow: str
    metadata: Dict[str, str]

    @classmethod
    def from_dict(cls, source_info: dict):
        filename = source_info['filename']
        flow = source_info['flow']
        metadata = source_info.get('metadata', {})
        return SourceInfo(filename=filename,
                          flow=flow,
                          metadata=metadata)

    def json(self):
        return {
            'filename': self.filename,
            'flow': self.flow,
            'metadata': self.metadata
        }


class DeltaFile(NamedTuple):
    did: str
    source_info: SourceInfo
    protocol_stack: List[ProtocolLayer]
    domains: List[Domain]
    indexed_metadata: Dict[str, str]
    enrichment: List[Domain]
    formatted_data: FormattedData

    @classmethod
    def from_dict(cls, delta_file: dict):
        did = delta_file['did']
        source_info = SourceInfo.from_dict(delta_file['sourceInfo'])
        protocol_stack = []
        if 'protocolStack' in delta_file:
            protocol_stack = [ProtocolLayer.from_dict(layer) for layer in delta_file['protocolStack']]
        domains = [Domain.from_dict(domain) for domain in delta_file['domains']]
        indexed_metadata = delta_file['indexedMetadata']
        enrichment = [Domain.from_dict(domain) for domain in delta_file['enrichment']]
        if len(delta_file['formattedData']) > 0:
            formatted_data = FormattedData.from_dict(delta_file['formattedData'][0])
        else:
            formatted_data = None
        return DeltaFile(did=did,
                         source_info=source_info,
                         protocol_stack=protocol_stack,
                         domains=domains,
                         indexed_metadata=indexed_metadata,
                         enrichment=enrichment,
                         formatted_data=formatted_data)


class Event(NamedTuple):
    delta_file: DeltaFile
    context: Context
    params: dict
    queue_name: str
    joined_delta_files: List[DeltaFile]
    return_address: str

    @classmethod
    def create(cls, event: dict, hostname: str, content_service: ContentService, logger: Logger):
        delta_file = DeltaFile.from_dict(event['deltaFile'])
        context = Context.create(event['actionContext'], hostname, content_service, logger)
        params = event['actionParams']
        queue_name = None
        if 'queueName' in event:
            queue_name = event['queueName']
        joined_delta_files = []
        if 'joinedDeltaFiles' in event:
            joined_delta_files = [DeltaFile.from_dict(joined_delta_file) for joined_delta_file in
                                  event['joinedDeltaFiles']]
        return_address = None
        if 'returnAddress' in event:
            return_address = event['returnAddress']
        return Event(delta_file, context, params, queue_name, joined_delta_files, return_address)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import io
import uuid
from typing import List, NamedTuple
from urllib.parse import urlparse

import minio

BUCKET = 'storage'


class Segment(NamedTuple):
    uuid: str
    offset: int
    size: int
    did: str

    def json(self):
        return {
            'uuid': str(self.uuid),
            'offset': self.offset,
            'size': self.size,
            'did': self.did
        }

    @classmethod
    def from_dict(cls, segment: dict):
        s_uuid = segment['uuid']
        offset = segment['offset']
        size = segment['size']
        did = segment['did']
        return Segment(uuid=s_uuid,
                       offset=offset,
                       size=size,
                       did=did)

    def id(self):
        return f"{self.did[:3]}/{self.did}/{self.uuid}"


class ContentReference(NamedTuple):
    segments: List[Segment]
    media_type: str

    def json(self):
        return {
            'segments': [segment.json() for segment in self.segments],
            'mediaType': self.media_type
        }

    @classmethod
    def from_dict(cls, content_reference: dict):
        segments = [Segment.from_dict(segment) for segment in content_reference['segments']]
        media_type = content_reference['mediaType']
        return ContentReference(segments=segments,
                                media_type=media_type)


class ContentService:
    def __init__(self, url, access_key, secret_key):
        parsed = urlparse(url)
        self.minio_client = minio.Minio(
            f"{parsed.hostname}:{str(parsed.port)}",
            access_key=access_key,
            secret_key=secret_key,
            secure=False
        )

        found = self.minio_client.bucket_exists(BUCKET)
        if not found:
            raise RuntimeError(f"Minio bucket {BUCKET} not found")

    def get_bytes(self, content_reference: ContentReference):
        return b"".join([self.minio_client.get_object(BUCKET, segment.id(), segment.offset,
                                                      segment.size).read() for segment in content_reference.segments])

    def get_str(self, content_reference: ContentReference):
        return self.get_bytes(content_reference).decode('utf-8')

    def put_bytes(self, did, bytes_data, media_type):
        segment = Segment(uuid=str(uuid.uuid4()),
                          offset=0,
                          size=len(bytes_data),
                          did=did)
        content_reference = ContentReference(segments=[segment],
                                             media_type=media_type)
        self.minio_client.put_object(BUCKET, segment.id(), io.BytesIO(bytes_data), len(bytes_data))
        return content_reference

    def put_str(self, did, string_data, media_type):
        return self.put_bytes(did, string_data.encode('utf-8'), media_type)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from datetime import datetime, timezone
from urllib.parse import urlparse

import redis
import time


class ActionEventQueue:
    def __init__(self, url, max_connections, password):
        parsed = urlparse(url)
        self.pool = redis.ConnectionPool(
            max_connections=max_connections,
            host=parsed.hostname,
            port=parsed.port,
            password=password)
        self.connection = None

    def get_connection(self):
        if self.connection is None:
            self.connection = redis.Redis(connection_pool=self.pool)
        return self.connection

    def size(self, name) -> int:
        conn = self.get_connection()
        count = conn.zcard(name)
        return count

    def put(self, name, item):
        now = round(time.time() * 1000)
        conn = self.get_connection()
        added = conn.zadd(name, {item: now}, nx=True)
        return added

    def take(self, name: str) -> str:
        conn = self.get_connection()
        setkey, item, score = conn.bzpopmin(name, 0)
        return item

    def heartbeat(self, name: str):
        conn = self.get_connection()
        utcnow = datetime.now(timezone.utc).isoformat().replace("+00:00", "Z")
        conn.hset("org.deltafi.action-queue.heartbeat", name, utcnow)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from abc import ABC, abstractmethod

from deltafi.actiontype import ActionType
from deltafi.domain import Context, DeltaFile
from deltafi.input import DomainInput, EgressInput, EnrichInput, FormatInput, LoadInput, TransformInput, ValidateInput
from deltafi.result import *
from pydantic import BaseModel


class Action(ABC):
    def __init__(self, action_type: ActionType, description: str, requires_domains: List[str],
                 requires_enrichments: List[str]):
        self.action_type = action_type
        self.description = description
        self.requires_domains = requires_domains
        self.requires_enrichments = requires_enrichments

    @abstractmethod
    def execute(self, event):
        pass

    def param_class(self):
        return BaseModel

    def validate_type(self, result, types: tuple):
        if not isinstance(result, types):
            raise ValueError(f"{self.__class__.__name__} must return one of "
                             f"{[result_type.__name__ for result_type in types]} "
                             f"but a {result.__class__.__name__} was returned")


class DomainAction(Action):
    def __init__(self, description: str, requires_domains: List[str]):
        super().__init__(ActionType.DOMAIN, description, requires_domains, [])

    def execute(self, event):
        domain_input = DomainInput(source_filename=event.delta_file.source_info.filename,
                                   ingress_flow=event.delta_file.source_info.flow,
                                   source_metadata=event.delta_file.source_info.metadata,
                                   metadata=event.delta_file.protocol_stack[-1].metadata,
                                   domains={domain.name: domain for domain in event.delta_file.domains})
        result = self.domain(event.context, self.param_class().parse_obj(event.params), domain_input)
        self.validate_type(result, (DomainResult, ErrorResult))
        return result

    @abstractmethod
    def domain(self, context: Context, params: BaseModel, domain_input: DomainInput):
        pass


class EgressAction(Action):
    def __init__(self, description: str):
        super().__init__(ActionType.EGRESS, description, [], [])

    def execute(self, event):
        egress_input = EgressInput(source_filename=event.delta_file.source_info.filename,
                                   ingress_flow=event.delta_file.source_info.flow,
                                   source_metadata=event.delta_file.source_info.metadata,
                                   formatted_data=event.delta_file.formatted_data)
        result = self.egress(event.context, self.param_class().parse_obj(event.params), egress_input)
        self.validate_type(result, (EgressResult, ErrorResult, FilterResult))
        return result

    @abstractmethod
    def egress(self, context: Context, params: BaseModel, egress_input: EgressInput):
        pass


class EnrichAction(Action):
    def __init__(self, description: str, requires_domains: List[str], requires_enrichments: List[str]):
        super().__init__(ActionType.ENRICH, description, requires_domains, requires_enrichments)

    def execute(self, event):
        enrich_input = EnrichInput(source_filename=event.delta_file.source_info.filename,
                                   ingress_flow=event.delta_file.source_info.flow,
                                   source_metadata=event.delta_file.source_info.metadata,
                                   content=event.delta_file.protocol_stack[-1].content,
                                   metadata=event.delta_file.protocol_stack[-1].metadata,
                                   domains={domain.name: domain for domain in event.delta_file.domains},
                                   enrichment={domain.name: domain for domain in event.delta_file.enrichment})
        result = self.enrich(event.context, self.param_class().parse_obj(event.params), enrich_input)
        self.validate_type(result, (EnrichResult, ErrorResult))
        return result

    @abstractmethod
    def enrich(self, context: Context, params: BaseModel, enrich_input: EnrichInput):
        pass


class FormatAction(Action):
    def __init__(self, description: str, requires_domains: List[str], requires_enrichments: List[str]):
        super().__init__(ActionType.FORMAT, description, requires_domains, requires_enrichments)

    def execute(self, event):
        format_input = FormatInput(source_filename=event.delta_file.source_info.filename,
                                   ingress_flow=event.delta_file.source_info.flow,
                                   source_metadata=event.delta_file.source_info.metadata,
                                   content=event.delta_file.protocol_stack[-1].content,
                                   metadata=event.delta_file.protocol_stack[-1].metadata,
                                   domains={domain.name: domain for domain in event.delta_file.domains},
                                   enrichment={domain.name: domain for domain in event.delta_file.enrichment})
        result = self.format(event.context, self.param_class().parse_obj(event.params), format_input)
        self.validate_type(result, (FormatResult, FormatManyResult, ErrorResult, FilterResult))
        return result

    @abstractmethod
    def format(self, context: Context, params: BaseModel, format_input: FormatInput):
        pass


class JoinAction(Action):
    def __init__(self, description: str):
        super().__init__(ActionType.JOIN, description, [], [])

    def execute(self, event):
        result = self.join(event.delta_file, event.joined_delta_files, event.context,
                           self.param_class().parse_obj(event.params))
        self.validate_type(result, (JoinResult, ErrorResult, FilterResult))
        return result

    @abstractmethod
    def join(self, delta_file: DeltaFile, joined_delta_files: List[DeltaFile], context: Context, params: BaseModel):
        pass


class LoadAction(Action):
    def __init__(self, description: str):
        super().__init__(ActionType.LOAD, description, [], [])

    def execute(self, event):
        load_input = LoadInput(source_filename=event.delta_file.source_info.filename,
                               ingress_flow=event.delta_file.source_info.flow,
                               source_metadata=event.delta_file.source_info.metadata,
                               content=event.delta_file.protocol_stack[-1].content,
                               metadata=event.delta_file.protocol_stack[-1].metadata)
        result = self.load(event.context, self.param_class().parse_obj(event.params), load_input)
        self.validate_type(result, (LoadResult, LoadManyResult, ErrorResult, FilterResult, SplitResult))
        return result

    @abstractmethod
    def load(self, context: Context, params: BaseModel, load_input: LoadInput):
        pass


class TransformAction(Action):
    def __init__(self, description: str):
        super().__init__(ActionType.TRANSFORM, description, [], [])

    def execute(self, event):
        transform_input = TransformInput(source_filename=event.delta_file.source_info.filename,
                                         ingress_flow=event.delta_file.source_info.flow,
                                         source_metadata=event.delta_file.source_info.metadata,
                                         content=event.delta_file.protocol_stack[-1].content,
                                         metadata=event.delta_file.protocol_stack[-1].metadata)
        result = self.transform(event.context, self.param_class().parse_obj(event.params), transform_input)
        self.validate_type(result, (TransformResult, ErrorResult, FilterResult))
        return result

    @abstractmethod
    def transform(self, context: Context, params: BaseModel, transform_input: TransformInput):
        pass


class ValidateAction(Action):
    def __init__(self, description: str):
        super().__init__(ActionType.VALIDATE, description, [], [])

    def execute(self, event):
        validate_input = ValidateInput(source_filename=event.delta_file.source_info.filename,
                                       ingress_flow=event.delta_file.source_info.flow,
                                       source_metadata=event.delta_file.source_info.metadata,
                                       formatted_data=event.delta_file.formatted_data)
        result = self.validate(event.context, self.param_class().parse_obj(event.params), validate_input)
        self.validate_type(result, (ValidateResult, ErrorResult, FilterResult))
        return result

    @abstractmethod
    def validate(self, context: Context, params: BaseModel, validate_input: ValidateInput):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['deltafi']

package_data = \
{'': ['*']}

install_requires = \
['json-logging==1.3.0',
 'minio==7.1.12',
 'pydantic==1.10.2',
 'redis==4.3.4',
 'requests==2.28.1',
 'urllib3==1.26.12']

setup_kwargs = {
    'name': 'deltafi',
    'version': '0.108.0',
    'description': 'SDK for DeltaFi plugins and actions',
    'long_description': '# DeltaFi Action Kit\n\nThis project provides a Python implementation of the DeltaFi Action Kit. The DeltaFi Action Kit is a setup of modules which simplify the creation of a DeltaFi Plugin.\n',
    'author': 'DeltaFi',
    'author_email': 'deltafi@systolic.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from enum import Enum


class ActionType(Enum):
    INGRESS = "ingress"
    TRANSFORM = "transform"
    LOAD = "load"
    JOIN = "join"
    DOMAIN = "domain"
    ENRICH = "enrich"
    FORMAT = "format"
    VALIDATE = "validate"
    EGRESS = "egress"
    DELETE = "delete"
    UNKNOWN = "unknown"



#
#    DeltaFi - Data transformation and enrichment platform
#
#    Copyright 2021-2023 DeltaFi Contributors <deltafi@deltafi.org>
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

import abc
from typing import Dict, List
import uuid

from deltafi.domain import Content, SourceInfo
from deltafi.metric import Metric
from deltafi.storage import ContentReference

ENDPOINT_TAG = "endpoint"
FILES_OUT = "files_out"
BYTES_OUT = "bytes_out"


class Result:
    __metaclass__ = abc.ABCMeta

    def __init__(self, result_key, result_type):
        self.result_key = result_key
        self.result_type = result_type
        self.metrics = []

    @abc.abstractmethod
    def response(self):
        pass

    def add_metric(self, metric: Metric):
        self.metrics.append(metric)


class DomainResult(Result):
    def __init__(self):
        super().__init__('domain', 'DOMAIN')
        self.indexed_metadata = {}

    def index_metadata(self, key: str, value: str):
        self.indexed_metadata[key] = value
        return self

    def response(self):
        return {
            'indexedMetadata': self.indexed_metadata
        }


class EgressResult(Result):
    def __init__(self, destination: str, bytes_egressed: int):
        super().__init__(None, 'EGRESS')
        self.add_metric(Metric(FILES_OUT, 1, {ENDPOINT_TAG: destination}))
        self.add_metric(Metric(BYTES_OUT, bytes_egressed, {ENDPOINT_TAG: destination}))

    def response(self):
        return None


class EnrichResult(Result):
    def __init__(self):
        super().__init__('enrich', 'ENRICH')
        self.enrichments = []
        self.indexed_metadata = {}

    def enrich(self, name: str, value: str, media_type: str):
        self.enrichments.append({
            'name': name,
            'value': value,
            'mediaType': media_type
        })
        return self

    def index_metadata(self, key: str, value: str):
        self.indexed_metadata[key] = value
        return self

    def response(self):
        return {
            'enrichments': self.enrichments,
            'indexedMetadata': self.indexed_metadata
        }


class ErrorResult(Result):
    def __init__(self, cause: str, context: str):
        super().__init__('error', 'ERROR')
        self.cause = cause
        self.context = context

    def response(self):
        return {
            'cause': self.cause,
            'context': self.context
        }


class FilterResult(Result):
    def __init__(self, filtered_cause: str):
        super().__init__('filter', 'FILTER')
        self.filtered_cause = filtered_cause

    def response(self):
        return {
            'message': self.filtered_cause
        }


class FormatResult(Result):
    def __init__(self, filename: str, content_reference: ContentReference):
        super().__init__('format', 'FORMAT')
        self.filename = filename
        self.content_reference = content_reference
        self.metadata = {}

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def response(self):
        return {
            'filename': self.filename,
            'contentReference': self.content_reference.json(),
            'metadata': self.metadata
        }


class FormatManyResult(Result):
    def __init__(self):
        super().__init__('formatMany', 'FORMAT_MANY')
        self.format_results = []

    def add_format_result(self, format_result: FormatResult):
        self.format_results.append(format_result)
        return self

    def response(self):
        return [format_result.response() for format_result in self.format_results]


class JoinResult(Result):
    def __init__(self, source_info: SourceInfo):
        super().__init__('join', 'JOIN')
        self.content = []
        self.metadata = {}
        self.source_info = source_info
        self.domains = []

    def add_content(self, content_list: List[Content]):
        self.content.extend(content_list)
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def add_domain(self, name: str, value: str, media_type: str):
        self.domains.append({
            'name': name,
            'value': value,
            'mediaType': media_type})
        return self

    def response(self):
        return {
            'sourceInfo': self.source_info.json(),
            'domains': self.domains,
            'protocolLayer': {
                'content': [content.json() for content in self.content],
                'metadata': self.metadata
            }
        }


class LoadResult(Result):
    def __init__(self):
        super().__init__('load', 'LOAD')
        self.content = []
        self.metadata = {}
        self.domains = []

    def add_content(self, name: str, content_reference: ContentReference):
        content = Content(name=name, content_reference=content_reference)
        self.content.append(content)
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def add_domain(self, name: str, value: str, media_type: str):
        self.domains.append({
            'name': name,
            'value': value,
            'mediaType': media_type})
        return self

    def response(self):
        return {
            'domains': self.domains,
            'protocolLayer': {
                'content': [content.json() for content in self.content],
                'metadata': self.metadata
            }
        }


class ChildLoadResult:
    def __init__(self, load_result: LoadResult = None):
        self._did = str(uuid.uuid4())
        self.load_result = load_result

    @property
    def did(self):
        return self._did

    def response(self):
        res = self.load_result.response()
        res["did"] = self._did
        return res


class LoadManyResult(Result):
    def __init__(self):
        super().__init__('loadMany', 'LOAD_MANY')
        self.load_results = []

    def add_load_result(self, load_result):
        if isinstance(load_result, ChildLoadResult):
            self.load_results.append(load_result)
        else:
            self.load_results.append(ChildLoadResult(load_result))
        return self

    def response(self):
        return [load_result.response() for load_result in self.load_results]


class SplitResult(Result):
    class SplitChild:
        def __init__(self, source_info: SourceInfo, content: List[Content]):
            self.source_info = source_info
            self.content = content

        def json(self):
            return {
                'sourceInfo': self.source_info.json(),
                'content': [content.json() for content in self.content]
            }

    def __init__(self):
        super().__init__('split', 'SPLIT')
        self.children = []

    def add_child(self, filename: str, flow: str, metadata: Dict[str, str], content: List[Content]):
        child = SplitResult.SplitChild(SourceInfo(filename, flow, metadata), content)
        self.children.append(child)

    def response(self):
        return [child.json() for child in self.children]


class TransformResult(Result):
    def __init__(self):
        super().__init__('transform', 'TRANSFORM')
        self.content = []
        self.metadata = {}

    def add_content(self, name: str, content_reference: ContentReference):
        content = Content(name=name, content_reference=content_reference)
        self.content.append(content)
        return self

    def add_metadata(self, key: str, value: str):
        self.metadata[key] = value
        return self

    def response(self):
        return {
            'protocolLayer': {
                'content': [content.json() for content in self.content],
                'metadata': self.metadata
            }
        }


class ValidateResult(Result):
    def __init__(self):
        super().__init__(None, 'VALIDATE')

    def response(self):
        return None
